#include "mysqlbackend.h"



/*
 *  Declaration of MySQL capabilities
 */

static struct odbx_basic_ops mysql_odbx_basic_ops = {
	.init = mysql_odbx_init,
	.bind = mysql_odbx_bind,
	.unbind = mysql_odbx_unbind,
	.finish = mysql_odbx_finish,
	.get_option = mysql_odbx_get_option,
	.set_option = mysql_odbx_set_option,
	.error = mysql_odbx_error,
	.error_type = mysql_odbx_error_type,
	.escape = mysql_odbx_escape,
	.query = mysql_odbx_query,
	.result = mysql_odbx_result,
	.result_free = mysql_odbx_result_free,
	.rows_affected = mysql_odbx_rows_affected,
	.row_fetch = mysql_odbx_row_fetch,
	.column_count = mysql_odbx_column_count,
	.column_name = mysql_odbx_column_name,
	.column_type = mysql_odbx_column_type,
	.field_length = mysql_odbx_field_length,
	.field_value = mysql_odbx_field_value,
};



struct odbx_ops mysql_odbx_ops = {
	.basic = &mysql_odbx_basic_ops
};



#ifndef ODBX_STATIC

void odbxdrv_register( struct odbx_ops** ops )
{
	*ops = &mysql_odbx_ops;
}

#endif



/*
 *  ODBX basic operations
 *  MySQL style
 */

static int mysql_odbx_init( odbx_t* handle, const char* host, const char* port )
{
	char* tmp = NULL;
	unsigned int portnum = 0;


	if( port != NULL )
	{
		portnum = (unsigned int) strtoul( port, &tmp, 10 );
		if( tmp[0] != '\0' )
		{
			return -ODBX_ERR_PARAM;
		}
	}

	if( ( handle->generic = malloc( sizeof( MYSQL ) ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	if( mysql_init( (MYSQL*) handle->generic ) == NULL )
	{
		free( handle->generic );
		handle->generic = NULL;

		return -ODBX_ERR_NOMEM;
	}

	if( ( handle->aux = malloc( sizeof( struct myconn ) ) ) == NULL )
	{
		free( handle->generic );
		handle->generic = NULL;

		return -ODBX_ERR_NOMEM;
	}

	struct myconn* aux = (struct myconn*) handle->aux;

	aux->port = portnum;
	aux->host = NULL;
	aux->mode = NULL;
	aux->flags = 0;
	aux->tls = 0;

	if( host != NULL )
	{
		size_t hlen = strlen( host ) + 1;

		if( ( aux->host = malloc( hlen ) ) == NULL )
		{
			free( handle->generic );
			free( handle->aux );

			handle->generic = NULL;
			handle->aux = NULL;

			return -ODBX_ERR_NOMEM;
		}

		memcpy( aux->host, host, hlen );
	}

	return ODBX_ERR_SUCCESS;
}



static int mysql_odbx_bind( odbx_t* handle, const char* database, const char* who, const char* cred, int method )
{
	struct myconn* param = (struct myconn*) handle->aux;

	if( handle->generic == NULL || param == NULL )
	{
		return -ODBX_ERR_PARAM;
	}

	if( method != ODBX_BIND_SIMPLE ) { return -ODBX_ERR_NOTSUP; }

	if( mysql_options( (MYSQL*) handle->generic, MYSQL_READ_DEFAULT_GROUP, "client" ) != 0 )
	{
		return -ODBX_ERR_BACKEND;
	}

	switch( param->tls )
	{
		case ODBX_TLS_TRY:

			param->flags |= CLIENT_SSL;

			if( mysql_real_connect( (MYSQL*) handle->generic, param->host,
				who, cred, database, param->port, NULL, param->flags ) != NULL )
			{
				goto SUCCESS;
			}

			param->flags &= ~CLIENT_SSL;
			break;

		case ODBX_TLS_ALWAYS:

			param->flags |= CLIENT_SSL;
			break;

		default:

			param->flags &= ~CLIENT_SSL;
	}

	if( mysql_real_connect( (MYSQL*) handle->generic, param->host,
		who, cred, database, param->port, NULL, param->flags ) == NULL )
	{
		return -ODBX_ERR_BACKEND;
	}

SUCCESS:

	return mysql_priv_setmode( handle, param->mode );
}



static int mysql_odbx_unbind( odbx_t* handle )
{
	if( handle->generic != NULL )
	{
		mysql_close( (MYSQL*) handle->generic );
		return ODBX_ERR_SUCCESS;
	}

	return -ODBX_ERR_PARAM;
}



static int mysql_odbx_finish( odbx_t* handle )
{
	struct myconn* aux = (struct myconn*) handle->aux;

	if( aux != NULL )
	{
		if( aux->host != NULL )
		{
			free( aux->host );
			aux->host = NULL;
		}

		free( handle->aux );
		handle->aux = NULL;
	}

	if( handle->generic != NULL )
	{
		mysql_thread_end();

		free( handle->generic );
		handle->generic = NULL;
	}

	return ODBX_ERR_SUCCESS;
}



static int mysql_odbx_get_option( odbx_t* handle, unsigned int option, void* value )
{
	switch( option )
	{
		case ODBX_OPT_API_VERSION:
			*((int*) value) = 0;
			break;
		case ODBX_OPT_THREAD_SAFE:
			*((int*) value) = (int) mysql_thread_safe();
			break;
		case ODBX_OPT_TLS:
			*((int*) value) = 1;   // FIXME: Howto find out if compiled with SSL support
			break;
		case ODBX_OPT_MULTI_STATEMENTS:
#ifdef HAVE_MYSQL_NEXT_RESULT
			*((int*) value) = 1;
#else
			*((int*) value) = 0;
#endif
			break;
		case ODBX_OPT_PAGED_RESULTS:
			*((int*) value) = 0;
			break;
		case ODBX_OPT_COMPRESS:
			*((int*) value) = 1;
			break;
		case ODBX_OPT_MODE:
			*((int*) value) = 1;
			break;
		default:
			return -ODBX_ERR_OPTION;
	}

	return ODBX_ERR_SUCCESS;
}



static int mysql_odbx_set_option( odbx_t* handle, unsigned int option, void* value )
{
	struct myconn* aux = (struct myconn*) handle->aux;

	if( handle->generic == NULL || aux == NULL )
	{
		return -ODBX_ERR_PARAM;
	}

	switch( option )
	{
		case ODBX_OPT_API_VERSION:
		case ODBX_OPT_THREAD_SAFE:
			return -ODBX_ERR_OPTRO;
		case ODBX_OPT_TLS:
			aux->tls = *((int*) value);
			return ODBX_ERR_SUCCESS;
		case ODBX_OPT_MULTI_STATEMENTS:
#ifdef HAVE_MYSQL_NEXT_RESULT
			if( *((int*) value) == 1 )
			{
				aux->flags |= CLIENT_MULTI_STATEMENTS;
				return ODBX_ERR_SUCCESS;
			}
			if( *((int*) value) == 0 )
			{
				aux->flags &= ~CLIENT_MULTI_STATEMENTS;
				return ODBX_ERR_SUCCESS;
			}
#endif
			break;
		case ODBX_OPT_PAGED_RESULTS:
			return ODBX_ERR_SUCCESS;
		case ODBX_OPT_COMPRESS:
			if( *((int*) value) == 1 )
			{
				aux->flags |= CLIENT_COMPRESS;
				return ODBX_ERR_SUCCESS;
			}
			if( *((int*) value) == 0 )
			{
				aux->flags &= ~CLIENT_COMPRESS;
				return ODBX_ERR_SUCCESS;
			}
			break;
		case ODBX_OPT_MODE:
			aux->mode = realloc( aux->mode, strlen( value ) + 1 );
			memcpy( aux->mode, value, strlen( value ) + 1 );
			return ODBX_ERR_SUCCESS;
		default:
			return -ODBX_ERR_OPTION;
	}

	return -ODBX_ERR_OPTWR;
}



static const char* mysql_odbx_error( odbx_t* handle )
{
	return mysql_error( (MYSQL*) handle->generic );
}



/*
 *  Server errors (1000-1999): non-fatal
 *  Client errors (2000-2999): fatal
 *
 *  This might be some kind of oversimplification
 *  but should be correct in most cases.
 */

static int mysql_odbx_error_type( odbx_t* handle )
{
	unsigned int errno = mysql_errno( (MYSQL*) handle->generic );

	if( !errno ) { return 0; }
	if( errno >= 1000 && errno < 2000 ) { return 1; }

	return -1;
}



static int mysql_odbx_escape( odbx_t* handle, const char* from, unsigned long fromlen, char* to, unsigned long* tolen )
{
	if( handle->generic == NULL )
	{
		return -ODBX_ERR_PARAM;
	}

	if( *tolen < fromlen * 2 + 1 )
	{
		return -ODBX_ERR_SIZE;
	}

	*tolen = mysql_real_escape_string( (MYSQL*) handle->generic, to, from, fromlen );

	return ODBX_ERR_SUCCESS;
}



static int mysql_odbx_query( odbx_t* handle, const char* query, unsigned long length )
{
	if( handle->generic == NULL || handle->aux == NULL )
	{
		return -ODBX_ERR_PARAM;
	}

	if( mysql_real_query( (MYSQL*) handle->generic, query, length ) != 0 )
	{
		return -ODBX_ERR_BACKEND;
	}

	((struct myconn*) handle->aux)->first = 1;
	return ODBX_ERR_SUCCESS;
}



static int mysql_odbx_result( odbx_t* handle, odbx_result_t** result, struct timeval* timeout, unsigned long chunk )
{
	MYSQL* conn = (MYSQL*) handle->generic;
	struct myconn* aux = (struct myconn*) handle->aux;


	if( conn == NULL || aux == NULL )
	{
		return -ODBX_ERR_PARAM;
	}

	if( aux->first == 0 )
	{
#ifdef HAVE_MYSQL_NEXT_RESULT
		switch( mysql_next_result( conn ) )
		{
			case -1:
				return 0; // no more results
			case 0:
				break;
			default:
				return -ODBX_ERR_BACKEND;
		}
#else
		return 0; // no more results
#endif
	}
	aux->first = 0;

	if( ( *result = (odbx_result_t*) malloc( sizeof( odbx_result_t ) ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	odbx_result_t* res = *result;
	res->generic = NULL;

	if( ( res->aux = malloc( sizeof( struct myres ) ) ) == NULL )
	{
		free( res );
		*result = NULL;

		return -ODBX_ERR_NOMEM;
	}

	struct myres* maux = (struct myres*) res->aux;

	if( ( maux->columns = (unsigned long) mysql_field_count( conn ) ) == 0 )
	{
		return 2;   // empty or not SELECT like query
	}

	if( chunk == 0 )
	{
		if( ( res->generic = (void*) mysql_store_result( conn ) ) == NULL )
		{
			free( res->aux );
			res->aux = NULL;

			free( res );
			*result = NULL;

			return -ODBX_ERR_BACKEND;
		}
	}
	else
	{
		if( ( res->generic = (void*) mysql_use_result( conn ) ) == NULL )
		{
			free( res->aux );
			res->aux = NULL;

			free( res );
			*result = NULL;

			return -ODBX_ERR_BACKEND;
		}
	}

	maux->fields = mysql_fetch_fields( (MYSQL_RES*) res->generic );

	return 3;   // result is available
}



static void mysql_odbx_result_free( odbx_result_t* result )
{
	if( result->generic != NULL )
	{
		mysql_free_result( (MYSQL_RES*) result->generic );
		result->generic = NULL;
	}

	if( result->aux != NULL )
	{
		free( result->aux );
		result->aux = NULL;
	}

	free( result );
}



static int mysql_odbx_row_fetch( odbx_result_t* result )
{
	MYSQL_RES* res = (MYSQL_RES*) result->generic;
	struct myres* aux = (struct myres*) result->aux;

	if( res == NULL || aux == NULL )
	{
		return -ODBX_ERR_PARAM;
	}

	if( ( aux->row = mysql_fetch_row( res ) ) == NULL )
	{
		aux->lengths = NULL;
		return 0;
	}

	aux->lengths = mysql_fetch_lengths( res );
	return 1;
}



static uint64_t mysql_odbx_rows_affected( odbx_result_t* result )
{
	my_ulonglong res;

	if( result->handle != NULL )
	{
		if( ( res = mysql_affected_rows( (MYSQL*) result->handle->generic ) ) != (my_ulonglong) -1 )
		{
			return (uint64_t) res;
		}
	}

	return 0;   // error
}



static unsigned long mysql_odbx_column_count( odbx_result_t* result )
{
	if( result->aux != NULL )
	{
		return ((struct myres*) result->aux)->columns;
	}

	return 0;
}



static const char* mysql_odbx_column_name( odbx_result_t* result, unsigned long pos )
{
	struct myres* aux = (struct myres*) result->aux;

	if( aux != NULL && aux->fields != NULL && pos < aux->columns )
	{
		return (const char*) aux->fields[pos].name;
	}

	return NULL;
}



static int mysql_odbx_column_type( odbx_result_t* result, unsigned long pos )
{
	struct myres* aux = (struct myres*) result->aux;

	if( aux != NULL && aux->fields != NULL && pos < aux->columns )
	{
		switch( aux->fields[pos].type )
		{
			case FIELD_TYPE_SHORT:
				return ODBX_TYPE_SMALLINT;
			case FIELD_TYPE_LONG:
				return ODBX_TYPE_INTEGER;
			case FIELD_TYPE_LONGLONG:
				return ODBX_TYPE_BIGINT;
			case FIELD_TYPE_DECIMAL:
				return ODBX_TYPE_DECIMAL;

			case FIELD_TYPE_FLOAT:
				return ODBX_TYPE_REAL;
			case FIELD_TYPE_DOUBLE:
				return ODBX_TYPE_DOUBLE;

			case FIELD_TYPE_STRING:
				return ODBX_TYPE_CHAR;
			case FIELD_TYPE_VAR_STRING:
				return ODBX_TYPE_VARCHAR;

			case FIELD_TYPE_TINY_BLOB:
			case FIELD_TYPE_MEDIUM_BLOB:
			case FIELD_TYPE_BLOB:
			case FIELD_TYPE_LONG_BLOB:
				if( aux->fields[pos].flags & BINARY_FLAG ) { return ODBX_TYPE_BLOB; }
				else { return ODBX_TYPE_CLOB; }

			case FIELD_TYPE_TIME:
				return ODBX_TYPE_TIME;
			case FIELD_TYPE_DATETIME:
				return ODBX_TYPE_TIMESTAMP;
			case FIELD_TYPE_DATE:
				return ODBX_TYPE_DATE;

			default:
				return ODBX_TYPE_UNKNOWN;
		}
	}

	return -ODBX_ERR_PARAM;
}



static unsigned long mysql_odbx_field_length( odbx_result_t* result, unsigned long pos )
{
	struct myres* aux = (struct myres*) result->aux;

	if( aux != NULL && aux->lengths != NULL && pos < aux->columns )
	{
		return aux->lengths[pos];
	}

	return 0;
}



static const char* mysql_odbx_field_value( odbx_result_t* result, unsigned long pos )
{
	struct myres* aux = (struct myres*) result->aux;

	if( aux != NULL && aux->row != NULL && pos < aux->columns )
	{
		return (const char*) aux->row[pos];
	}

	return NULL;
}



/*
 * MySQL private functions
 */

static int mysql_priv_setmode( odbx_t* handle, const char* mode )
{
	char* stmt;
	char* lmode = "ANSI";
	size_t modelen = 4;
	size_t len = 24;


	if( mode != NULL )
	{
		modelen = strlen( mode );
		lmode = (char*) mode;
	}

	if( ( stmt = (char*) malloc( len + modelen ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	memcpy( stmt, "SET SESSION sql_mode='", 22 );
	memcpy( stmt + 22, lmode, modelen );
	memcpy( stmt + 22 + modelen, "'", 2 );

	if( mysql_real_query( (MYSQL*) handle->generic, stmt, len + modelen ) != 0 )
	{
		return -ODBX_ERR_BACKEND;
	}

	if( mysql_field_count( (MYSQL*) handle->generic ) != 0 )
	{
		MYSQL_RES* result;

		if( ( result = mysql_store_result( (MYSQL*) handle->generic ) ) == NULL )
		{
			return -ODBX_ERR_BACKEND;
		}

		mysql_free_result( result );
	}

	free( stmt );

	return ODBX_ERR_SUCCESS;
}
