#include <stdlib.h>
#include "lib/odbx.h"
#include "lib/odbxlib.h"
#include "lib/odbxdrv.h"



static const char* odbx_errmsg[] = {
	gettext_noop("Success"),
	gettext_noop("Backend error occured"),
	gettext_noop("Invalid capability"),
	gettext_noop("Invalid parameter"),
	gettext_noop("Out of memory"),
	gettext_noop("Incorrect size of allocated memory"),
	gettext_noop("Loading backend library failed"),
	gettext_noop("Operation is not available"),
	gettext_noop("Invalid option"),
	gettext_noop("Option is read only"),
	gettext_noop("Setting option failed"),
	gettext_noop("Waiting for result failed"),
	gettext_noop("Not supported"),
};



int odbx_init( odbx_t** handle, const char* backend, const char* host, const char* port )
{
	int err;


	if( handle == NULL || backend == NULL )
	{
		return -ODBX_ERR_PARAM;
	}

	if( ( *handle = malloc( sizeof( struct odbx_t ) ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	(*handle)->ops = NULL;
	(*handle)->backend = NULL;
	(*handle)->generic = NULL;
	(*handle)->aux = NULL;

	if( ( err = _odbx_lib_open( *handle, backend ) ) < 0 )
	{
		free( *handle );
		*handle = NULL;
		return err;
	}

	if( (*handle)->ops && (*handle)->ops->basic && (*handle)->ops->basic->init )
	{
		if( ( err = (*handle)->ops->basic->init( *handle, host, port ) ) < 0 )
		{
			_odbx_lib_close( *handle );

			free( *handle );
			*handle = NULL;
		}

		return err;
	}

	_odbx_lib_close( *handle );

	free( *handle );
	*handle = NULL;

	return -ODBX_ERR_NOOP;
}



int odbx_bind( odbx_t* handle, const char* database, const char* who, const char* cred, int method )
{
	if( database != NULL && handle != NULL && handle->ops != NULL && handle->ops->basic != NULL && handle->ops->basic->bind != NULL )
	{
		return handle->ops->basic->bind( handle, database, who, cred, method );
	}

	return -ODBX_ERR_PARAM;
}



/* Depricated: odbx_bind_simple() */

int odbx_bind_simple( odbx_t* handle, const char* database, const char* username, const char* password )
{
	return odbx_bind( handle, database, username, password, ODBX_BIND_SIMPLE );
}



int odbx_unbind( odbx_t* handle )
{
	if( handle != NULL && handle->ops != NULL && handle->ops->basic != NULL && handle->ops->basic->unbind != NULL )
	{
		return handle->ops->basic->unbind( handle );
	}

	return -ODBX_ERR_PARAM;
}



int odbx_finish( odbx_t* handle )
{
	if( handle != NULL && handle->ops != NULL && handle->ops->basic != NULL && handle->ops->basic->finish != NULL )
	{
		int err;

		if( ( err = handle->ops->basic->finish( handle ) ) < 0 )
		{
			return err;
		}

		if( ( err = _odbx_lib_close( handle ) ) < 0 )
		{
			return err;
		}

		handle->ops = NULL;
		free( handle );

		return ODBX_ERR_SUCCESS;
	}

	return -ODBX_ERR_PARAM;
}



int odbx_capabilities( odbx_t* handle, unsigned int cap )
{
	if( handle != NULL && handle->ops != NULL )
	{
		switch( cap )
		{
			case ODBX_CAP_BASIC:
				if( handle->ops->basic != NULL ) { return 1; }
		}
		return 0;
	}

	return -ODBX_ERR_PARAM;
}



int odbx_get_option( odbx_t* handle, unsigned int option, void* value )
{
	if( value != NULL && handle != NULL && handle->ops != NULL && handle->ops->basic != NULL && handle->ops->basic->get_option != NULL )
	{
		return handle->ops->basic->get_option( handle, option, value );
	}

	return -ODBX_ERR_PARAM;
}



int odbx_set_option( odbx_t* handle, unsigned int option, void* value )
{
	if( value != NULL && handle != NULL && handle->ops != NULL && handle->ops->basic != NULL && handle->ops->basic->set_option != NULL )
	{
		return handle->ops->basic->set_option( handle, option, value );
	}

	return -ODBX_ERR_PARAM;
}



const char* odbx_error( odbx_t* handle, int error )
{
	if( error == -ODBX_ERR_BACKEND )
	{
		if( handle != NULL && handle->ops != NULL && handle->ops->basic != NULL && handle->ops->basic->error != NULL )
		{
			return handle->ops->basic->error( handle );
		}

		return dgettext( "opendbx", odbx_errmsg[ODBX_ERR_PARAM] );
	}

	if( error <= ODBX_ERR_SUCCESS && error > -ODBX_MAX_ERRNO )
	{
		return dgettext( "opendbx", odbx_errmsg[-error] );
	}

	return dgettext( "opendbx", odbx_errmsg[ODBX_ERR_PARAM] );
}



int odbx_error_type( odbx_t* handle, int errno )
{
	switch( errno )
	{
		case ODBX_ERR_SUCCESS:

			return 0;

		case -ODBX_ERR_BACKEND:

			if( handle != NULL && handle->ops != NULL && handle->ops->basic != NULL && handle->ops->basic->error_type != NULL )
			{
				return handle->ops->basic->error_type( handle );
			}

		case -ODBX_ERR_NOMEM:
		case -ODBX_ERR_NOTEXIST:
		case -ODBX_ERR_NOOP:
		case -ODBX_ERR_RESULT:

			return -1;
	}

	return 1;
}



int odbx_escape( odbx_t* handle, const char* from, unsigned long fromlen, char* to, unsigned long* tolen )
{
	if( from != NULL && to != NULL && tolen != NULL && handle != NULL && handle->ops != NULL && handle->ops->basic != NULL )
	{
		if(  handle->ops->basic->escape )
		{
			return handle->ops->basic->escape( handle, from, fromlen, to, tolen );
		}

		unsigned long i, len = 0;

		for( i = 0; i < fromlen; i++ )
		{
			if( i == *tolen - 1 ) { return -ODBX_ERR_SIZE; }

			if( from[i] == '\'' ) { to[len++] = '\''; }

			to[len++] = from[i];
		}

		to[len] = '\0';
		*tolen = len;

		return ODBX_ERR_SUCCESS;
	}

	return -ODBX_ERR_PARAM;
}



int odbx_query( odbx_t* handle, const char* query, unsigned long length )
{
	if( handle != NULL && handle->ops != NULL && handle->ops->basic != NULL && handle->ops->basic->query != NULL )
	{
		return handle->ops->basic->query( handle, query, length );
	}

	return -ODBX_ERR_PARAM;
}



int odbx_result( odbx_t* handle, odbx_result_t** result, struct timeval* timeout, unsigned long chunk )
{
	if( handle != NULL && handle->ops != NULL && handle->ops->basic != NULL && handle->ops->basic->result != NULL )
	{
		int err;
		*result = NULL;

		if( ( err = handle->ops->basic->result( handle, result, timeout, chunk ) ) > 1 )
		{
			if( *result ) { (*result)->handle = handle; }
		}

		return err;
	}

	return -ODBX_ERR_PARAM;
}



void odbx_result_free( odbx_result_t* result )
{
	if( result != NULL && result->handle != NULL && result->handle->ops != NULL &&
		result->handle->ops->basic != NULL && result->handle->ops->basic->result_free != NULL )
	{
		result->handle->ops->basic->result_free( result );
	}
}



int odbx_row_fetch( odbx_result_t* result )
{
	if( result != NULL && result->handle != NULL && result->handle->ops != NULL &&
		result->handle->ops->basic != NULL && result->handle->ops->basic->row_fetch != NULL )
	{
		return result->handle->ops->basic->row_fetch( result );
	}

	return -ODBX_ERR_PARAM;
}



uint64_t odbx_rows_affected( odbx_result_t* result )
{
	if( result != NULL && result->handle != NULL && result->handle->ops != NULL &&
		result->handle->ops->basic != NULL && result->handle->ops->basic->rows_affected != NULL )
	{
		return result->handle->ops->basic->rows_affected( result );
	}

	return 0;
}



unsigned long odbx_column_count( odbx_result_t* result )
{
	if( result != NULL && result->handle != NULL && result->handle->ops != NULL &&
		result->handle->ops->basic != NULL && result->handle->ops->basic->column_count != NULL )
	{
		return result->handle->ops->basic->column_count( result );
	}

	return 0;
}



const char* odbx_column_name( odbx_result_t* result, unsigned long pos )
{
	if( result != NULL && result->handle != NULL && result->handle->ops != NULL &&
		result->handle->ops->basic != NULL && result->handle->ops->basic->column_name != NULL )
	{
		return result->handle->ops->basic->column_name( result, pos );
	}

	return NULL;
}



int odbx_column_type( odbx_result_t* result, unsigned long pos )
{
	if( result != NULL && result->handle != NULL && result->handle->ops != NULL &&
		result->handle->ops->basic != NULL && result->handle->ops->basic->column_type != NULL )
	{
		return result->handle->ops->basic->column_type( result, pos );
	}

	return -ODBX_ERR_PARAM;
}



unsigned long odbx_field_length( odbx_result_t* result, unsigned long pos )
{
	if( result != NULL && result->handle != NULL && result->handle->ops != NULL &&
		result->handle->ops->basic != NULL && result->handle->ops->basic->field_length != NULL )
	{
		return result->handle->ops->basic->field_length( result, pos );
	}

	return 0;
}



const char* odbx_field_value( odbx_result_t* result, unsigned long pos )
{
	if( result != NULL && result->handle != NULL && result->handle->ops != NULL &&
		result->handle->ops->basic != NULL && result->handle->ops->basic->field_value != NULL )
	{
		return result->handle->ops->basic->field_value( result, pos );
	}

	return NULL;
}
