/*
 *  OpenDBX - A simple but extensible database abstraction layer
 *  Copyright (C) 2004-2006 Norbert Sendetzky <norbert@linuxnetworks.de>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA.
 */



#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STDINT_H
#include <stdint.h>
#else
#include <sys/types.h>
#endif

#include <sys/time.h>



#ifndef ODBX_H
#define ODBX_H



#ifdef __cplusplus
extern "C" {
#endif



/*
 *  ODBX (SQL2003) data types
 */

#define ODBX_TYPE_BOOLEAN   0x00
#define ODBX_TYPE_SMALLINT   0x01
#define ODBX_TYPE_INTEGER   0x02
#define ODBX_TYPE_BIGINT   0x03
#define ODBX_TYPE_DECIMAL   0x07

#define ODBX_TYPE_REAL   0x08
#define ODBX_TYPE_DOUBLE   0x09
#define ODBX_TYPE_FLOAT   0x0f

#define ODBX_TYPE_CHAR   0x10
#define ODBX_TYPE_NCHAR   0x11
#define ODBX_TYPE_VARCHAR   0x12
#define ODBX_TYPE_NVARCHAR  0x13

#define ODBX_TYPE_CLOB   0x20
#define ODBX_TYPE_NCLOB   0x21
#define ODBX_TYPE_XML   0x22

#define ODBX_TYPE_BLOB   0x2f

#define ODBX_TYPE_TIME   0x30
#define ODBX_TYPE_TIMETZ   0x31
#define ODBX_TYPE_TIMESTAMP   0x32
#define ODBX_TYPE_TIMESTAMPTZ 0x33
#define ODBX_TYPE_DATE   0x34
#define ODBX_TYPE_INTERVAL   0x35

#define ODBX_TYPE_ARRAY   0x40
#define ODBX_TYPE_MULTISET   0x41

#define ODBX_TYPE_DATALINK   0x50

#define ODBX_TYPE_UNKNOWN   0xff



/*
 *  Extended capabilities supported by the backends
 *  0x0000-0x00ff: Well known capabilities
 */

#define ODBX_CAP_BASIC   0x0000



/*
 *  ODBX error types
 */

#define ODBX_ERR_SUCCESS   0x00
#define ODBX_ERR_BACKEND   0x01
#define ODBX_ERR_NOCAP   0x02
#define ODBX_ERR_PARAM   0x03
#define ODBX_ERR_NOMEM   0x04
#define ODBX_ERR_SIZE   0x05
#define ODBX_ERR_NOTEXIST   0x06
#define ODBX_ERR_NOOP   0x07
#define ODBX_ERR_OPTION   0x08
#define ODBX_ERR_OPTRO   0x09
#define ODBX_ERR_OPTWR   0x0a
#define ODBX_ERR_RESULT   0x0b
#define ODBX_ERR_NOTSUP   0x0c

#define ODBX_MAX_ERRNO   0x0c



/*
 * ODBX bind type
 */

#define ODBX_BIND_SIMPLE   0



/*
 *  ODBX options
 *
 *  0x0000 - 0x1fff reserved for api options
 *  0x2000 - 0x3fff reserved for api extension options
 *  0x4000 - 0xffff reserved for vendor specific and experimental options
 */

#define ODBX_DISABLE   0
#define ODBX_ENABLE   1

/*
 *  Informational options
 */

#define ODBX_OPT_API_VERSION   0x0000
#define ODBX_OPT_THREAD_SAFE   0x0001


/*
 *  Security related options
 */

#define ODBX_TLS_NEVER   0x00
#define ODBX_TLS_TRY   0x01
#define ODBX_TLS_ALWAYS   0x02

#define ODBX_OPT_TLS   0x0010


/*
 *  Implemented options
 */

#define ODBX_OPT_MULTI_STATEMENTS   0x0020
#define ODBX_OPT_PAGED_RESULTS   0x0021
#define ODBX_OPT_COMPRESS   0x0022
#define ODBX_OPT_MODE   0x0023



typedef struct odbx_t odbx_t;
typedef struct odbx_result_t odbx_result_t;



/*
 *  ODBX basic operations
 */

int odbx_init( odbx_t** handle, const char* backend, const char* host, const char* port );

int odbx_bind( odbx_t* handle, const char* database, const char* who, const char* cred, int method );

int odbx_unbind( odbx_t* handle );

int odbx_finish( odbx_t* handle );

int odbx_capabilities( odbx_t* handle, unsigned int cap );

int odbx_get_option( odbx_t* handle, unsigned int option, void* value );

int odbx_set_option( odbx_t* handle, unsigned int option, void* value );

const char* odbx_error( odbx_t* handle, int error );

int odbx_error_type( odbx_t* handle, int error );

int odbx_escape( odbx_t* handle, const char* from, unsigned long fromlen, char* to, unsigned long* tolen );

int odbx_query( odbx_t* handle, const char* query, unsigned long length );

int odbx_result( odbx_t* handle, odbx_result_t** result, struct timeval* timeout, unsigned long chunk );

void odbx_result_free( odbx_result_t* result );

int odbx_row_fetch( odbx_result_t* result );

uint64_t odbx_rows_affected( odbx_result_t* result );

unsigned long odbx_column_count( odbx_result_t* result );

const char* odbx_column_name( odbx_result_t* result, unsigned long pos );

int odbx_column_type( odbx_result_t* result, unsigned long pos );

unsigned long odbx_field_length( odbx_result_t* result, unsigned long pos );

const char* odbx_field_value( odbx_result_t* result, unsigned long pos );



/*
 *  Depricated defines and functions
 *
 *  They won't be available in version 2.0 any more
 *  Please don't use and replace them in your code
 */


/*
 * Depricated types
 */

#define ODBX_BOOLEAN   0x00
#define ODBX_SMALLINT   0x01
#define ODBX_INTEGER   0x02
#define ODBX_BIGINT   0x03
#define ODBX_DECIMAL   0x07

#define ODBX_REAL   0x08
#define ODBX_DOUBLE   0x09
#define ODBX_FLOAT   0x0f

#define ODBX_CHAR   0x10
#define ODBX_NCHAR   0x11
#define ODBX_VARCHAR   0x12
#define ODBX_NVARCHAR  0x13

#define ODBX_CLOB   0x20
#define ODBX_NCLOB   0x21
#define ODBX_XML   0x22

#define ODBX_BLOB   0x2f

#define ODBX_TIME   0x30
#define ODBX_TIME_TZ   0x31
#define ODBX_TIMESTAMP   0x32
#define ODBX_TIMESTAMP_TZ 0x33
#define ODBX_DATE   0x34
#define ODBX_INTERVAL   0x35

#define ODBX_ARRAY   0x40
#define ODBX_MULTISET   0x41

#define ODBX_DATALINK   0x50

#define ODBX_UNKNOWN   0xff


/*
 * Depricated errors
 */

#define ODBX_ERR_TOOLONG ODBX_ERR_SIZE


/*
 * Depricated functions
 */

int odbx_bind_simple( odbx_t* handle, const char* database, const char* username, const char* password );



#ifdef  __cplusplus
}
#endif


#endif
